;;; beg-end.el

;; Author: Andreas Roehler
;; <andreas.roehler@online.de>

;; Keywords: languages

;; This file is free software; you can redistribute it
;; and/or modify it under the terms of the GNU General
;; Public License as published by the Free Software
;; Foundation; either version 2, or (at your option)
;; any later version.
;; This file is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the
;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;; PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.
;; You should have received a copy of the GNU General
;; Public License along with GNU Emacs; see the file
;; COPYING.  If not, write to the Free Software
;; Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Major changes to previous version: 
;;  Checks for inside comment deactivated, circle-call-bugs

;;; Commentary:

;; Routine detecting nested forms - for all kind of
;; things which have a start- and a end-string.
;; End-string might be the empty string.  Jump to the
;; beginning or end of a form.  Repeated execution
;; forward or backward.  With argument as many times.

(require 'misc-utils)

;;; Code:

;; (setq begstr "\\bcase\\b\\|\\bfor\\b\\|\\bfunction\\b\\|\\bif\\b\\|\\bselect\\b\\|\\buntil\\b\\|\\bwhile\\b")
;; (setq endstr "\\bdone\\b\\|\\besac\\b\\|\\bfi\\b") 

(setq match-paren-char "%")

;; (defun set-beginning-of-form ()
;;   "Set `beginning-of-form' as `beginning-of-defun-function'."

;; (defun set-end-of-form ()
;;   "Set `end-of-form' as `end-of-defun-function'."
;;   (interactive)
;;   (set (make-local-variable 'end-of-defun-function) 'end-of-form))

(defun mark-form (begstr endstr &optional bound noerror count comment)
  (beginning-of-form begstr endstr bound noerror count comment)
  (push-mark (point) t t)
  (end-of-form begstr endstr bound noerror count comment)
  (kill-new (buffer-substring-no-properties (mark) (point))))

(defun kill-form (begstr endstr &optional bound noerror count comment)
  (beginning-of-form begstr endstr bound noerror count comment)
  (push-mark (point) t t)
  (end-of-form begstr endstr bound noerror count comment)
  (kill-region (mark) (point)))

(defun beg-end-match-paren-base (begstr endstr &optional iact)
  "Switch between opening and closing programming constructs.
If not at the beginning or the end, to opening first. "
  (cond ((ignore-errors (looking-at begstr))
	 (end-of-form))
	((progn (backward-word)
		(ignore-errors (looking-at endstr)))
	 (beginning-of-form)
	 (when iact (message "%s" (point))) (point))
	(t (beginning-of-form)
	   (when iact (message "%s" (point))) (point))))

(defun beginning-of-form (&optional bound noerror count comment) 
  "Goto opening of a programming structure in this level.
Reads strings as arguments from minibuffer.
 Don't use this from a program, use `beginning-of-form-base' instead. "
  (interactive)
  (let ((begstr (read-from-minibuffer "Insert start: "))
	(endstr (read-from-minibuffer "Insert endstr: ")))
    (beginning-of-form-base begstr endstr bound noerror count comment))
  (when iact (message "%s" (point))) (point))

(defun end-of-form (&optional iact bound noerror count comment)
  "Goto opening of a programming structure in this level.
Reads strings as arguments from minibuffer.
Set comment to `t' if forms inside comments should match - also for processing comments itself.
 Don't use this from a program, use `end-of-form-base' instead. "
  (interactive "p")
  (let ((begstr (read-from-minibuffer "Insert start: "))
	(endstr (read-from-minibuffer "Insert endstr: ")))
    (end-of-form-base begstr endstr bound noerror count comment))
  (when iact (message "%s" (point))) (point))

(defun ar-leave-begstr-backward (begstr unquoted-beg)
  (let* ((stringcount (length unquoted-beg))
         (collected (char-to-string (char-after)))
         (indx (string-match (regexp-quote collected) unquoted-beg)))
    (while (and indx (not (ignore-errors (looking-at begstr)))(< 1 stringcount))
      (forward-char -1)
      (setq collected (concat (char-to-string (char-after)) collected))
      (setq indx (string-match (regexp-quote collected) unquoted-beg))
      (setq stringcount (1- stringcount)))))

(defun ar-leave-endstr-backward (endstr unquoted-end)
  (let* ((stringcount (length unquoted-end))
         (collected (char-to-string (char-after)))
         (indx (string-match (regexp-quote collected) unquoted-end)))
    (while (and indx (not (ignore-errors (looking-at endstr)))(< 1 stringcount))
      (forward-char -1)
      (setq collected (concat (char-to-string (char-after)) collected))
      (setq indx (string-match (regexp-quote collected) unquoted-end))
      (setq stringcount (1- stringcount)))))

;; should `parse-sexp-ignore-comments' be used?
(defun beginning-of-form-base (begstr &optional endstr bound noerror count comment regexp-quote regexp condition) 
  "Goto opening of a programming structure in this level.
Takes strings as arguments for search.
Set comment to `t' if forms inside comments should match - also for processing comments itself.
Set REGEXP-QUOTE to t, if end-form should be quoted.
Set 8th argument REGEXP t, if beg/end-str are regular expressions.
"
  (lexical-let* ((nesting (if count count 1))
                 (beg-end-unequal (not (string= begstr endstr)))
                 (begstr (if regexp-quote (regexp-quote begstr) begstr))
                 (endstr (cond ((and (stringp endstr)(string= endstr ""))
                                nil)
                               (regexp-quote (regexp-quote endstr)) 
                               (t endstr)))
                 (searchform (cond ((and beg-end-unequal begstr endstr)
                                    (concat "\\(" begstr "\\)\\|\\(" endstr "\\)"))
                                   (begstr begstr)
                                   (endstr endstr)))
                 (orig (point))
                 (bound (or bound (if regexp (line-beginning-position)(- (point)(length begstr)))))
                 beg-pos-delimiter end-pos-delimiter)
    (while
        (and (< 0 nesting) (not (bobp)))
      (if (or regexp beg-end-unequal)
          (or (re-search-backward searchform nil t 1)
              ;; re-search-backward may start inside searched
              (while (and bound (ignore-errors (< bound (point)))(not (looking-at begstr)))(forward-char -1)))
        (or (and (search-backward searchform nil t 1)(not (and condition (funcall condition))))
            ;; search-backward may start inside searched
            (while (and bound (ignore-errors (< bound (point)))(not (looking-at begstr)))(forward-char -1))))
      (if (< (point) orig)
          (cond ((and condition (funcall condition))
                 (beginning-of-form-base begstr endstr nil noerror count comment regexp-quote regexp condition)) 
                ((and (looking-at begstr))
                 (setq beg-pos-delimiter (match-beginning 0))
                 (setq end-pos-delimiter (match-end 0))
                 (if comment
                     (setq nesting (1- nesting))
                   (unless (ar-in-comment-p-lor)
                     (setq nesting (1- nesting)))))
                ((string-match endstr (match-string-no-properties 0))
                 ;; another endstring found
                 (if comment
                     ;; non-nesting comments don't matter
                     (unless (or (string= "" comment-end)(eq 10 comment-end))
                       (setq nesting (1+ nesting)))
                   (unless (ar-in-comment-p-lor) 
                     (setq nesting (1+ nesting))))))
        ;; not success at all 
        (setq nesting 0)))
    (if (and beg-pos-delimiter end-pos-delimiter)
        (list beg-pos-delimiter end-pos-delimiter)
      (goto-char (point-min)))))

(defun end-of-form-base (begstr &optional endstr bound noerror count comment regexp-quote regexp condition)
  "Goto closing of a programming structure in this level.
As it stops one char after form, go one char back onto the last char of form.
Set comment to `t' if forms inside comments should match - also for processing comments itself.
If SHOW, display nesting and point in message buffer.
Set REGEXP-QUOTE to t, if end-form should be quoted.
Set 8th argument REGEXP t, if beg/end-strings are regular expressions.
Optional arg CONDITION expects a function whose return value - `t' or a number - is checked for success, otherwise search continues.
"
  (lexical-let* ((unquoted-end endstr)
                 (beg-end-unequal (not (string= begstr endstr)))
                 (begstr (cond
                          ((string= begstr "")
                           nil)
                          (regexp-quote (regexp-quote begstr))
                          (t begstr)))
                 (endstr (cond ((string= endstr "")
                                nil)
                               (regexp-quote (regexp-quote endstr)) 
                               (t endstr)))
                 (searchform (cond ((and beg-end-unequal begstr endstr)
                                    (concat "\\(" begstr "\\)\\|\\(" endstr "\\)"))
                                   (begstr begstr)
                                   (endstr endstr)))
                 (nesting (or count count 1))
                 (orig (point)) 
                 ;; if inside endstr, go to the end
                 beg-pos-delimiter end-pos-delimiter)
    (if (ignore-errors (looking-at begstr))
        (goto-char (match-end 0))
      ;; in inside endstring, go backward to its beginning
      (ar-leave-endstr-backward endstr unquoted-end))
    (while
        (and
         (< 0 nesting) (not (eobp))
         (re-search-forward searchform bound noerror 1))
      (setq beg-pos-delimiter (match-beginning 0))
      (setq end-pos-delimiter (match-end 0))
      (if
          (string-match endstr (match-string-no-properties 0))
          (if comment
              (setq nesting (1- nesting))
            (unless (or (ar-in-comment-p-lor)(and condition (funcall condition)))
              (setq nesting (1- nesting))))
        ;; got another beginning while moving down
        (if comment
            (setq nesting (1+ nesting))
          (unless (or (ar-in-comment-p-lor) (and condition (funcall condition)))
            (setq nesting (1+ nesting))))))
    (if (and beg-pos-delimiter end-pos-delimiter)
        (list beg-pos-delimiter end-pos-delimiter)
      (goto-char orig))))

(defvar match-paren-key-char "%")

(defvar be-match-paren-mode nil)

(defun be-match-paren-mode (&optional iact) 
  "Toggle be-match-paren-mode. 
If on, inserting of `be-match-paren-char', default is \"%\", moves to the matching opening/closing.
With arg, insert the charakter the key is on 
Key per default is \"%\" as with elisp's `match-paren'. "
  (interactive "p")
  (if be-match-paren-mode
      (progn 
	(setq be-match-paren-mode nil)
	(define-key be-mode-map "%" 'self-insert-command)
	(when iact (message "be-match-paren-mode: %s" be-match-paren-mode)))
    (setq be-match-paren-mode t)
    (define-key be-mode-map "%" 'be-match-paren)
    (when iact (message "be-match-paren-mode: %s" be-match-paren-mode))))

(defun beg-end-match-paren (begstr endstr &optional ins) 
  "Go to the matching opening/closing. 
First to opening, unless cursor is already there.
With arg, insert the charakter of `sh-match-paren-char'.
Key per default is \"%\" as with elisp's `match-paren'. "
  (if ins
      (insert match-paren-key-char)
    (cond ((ignore-errors (looking-at begstr))
	   (end-of-form-base begstr endstr bound noerror count))
	  ((ignore-errors (looking-at "\\s("))
	   (match-paren arg))
	  ((ignore-errors (looking-at "\\s)"))
	   (match-paren arg))
	  (t (beginning-of-form-base begstr endstr bound noerror count comment)))))

(defun in-delimited-form-intern (form &optional regexp)
  "Returns beginning of string if inside, 0 otherwise"
  (save-excursion
    (let* ((pmin (point-min))
           in-string
           (end (or (progn
                      (cond
                       ((looking-at (substring form -3))
                        (setq in-string 3)
                        (match-end 0))
                       ((looking-at (substring form -2))
                        (setq in-string 2)
                        (match-end 0))
                       ((looking-at (substring form -1))
                        (setq in-string 1)
                        (match-end 0))))
                    (point)))
           (count 0)
           (before '())
           erg done)
      (save-restriction
        (widen)
        (goto-char (point-min))
        (when (and (bobp) 
                   (looking-at form))
          (setq count (1+ count))
          (goto-char (match-end 0)))
        (unless (< end (point))
          (while
              (if regexp
                  (setq erg (re-search-forward form end t 1))
                (setq erg (search-forward form end t 1)))
            (unless (ar-in-comment-p-lor) 
              (setq count (1+ count))
              (add-to-list 'before (match-beginning 0)))))
        (when (and (not (eq 1 (% count 2))) (numberp in-string))
          (forward-char (- in-string))
          (while (or (looking-at (substring form -3))
                     (looking-at (substring form -2))
                     (looking-at (substring form -1)))
            (forward-char -1))
          (when (and (looking-at form)(not (ar-in-comment-p-lor)))
            (setq count (1- count)))
          (setq erg (nth 1 before)) (setq done t))
        (when (eq 1 (% count 2))
          (unless done
            (setq erg (match-beginning 0))))
        (when iact (message "%s" erg))
        erg))))

(defun in-form-base (begstr endstr &optional beg end count comment regexp-quote regexp condition) 
  "Parse a programming structure. Return level of nesting, `nil' if not inside.
Takes strings as arguments for search.
Set comment to `t' if forms inside comments should match - also for processing comments itself.
Set REGEXP-QUOTE to t, if end-form should be quoted.
Set 8th argument REGEXP t, if beg/end-str are regular expressions.
"
  (let* ((beg (or beg (point-min)))
         (end (or end (point)))
         (nesting (or count count 0))
         (unquoted-end endstr)
         (beg-end-unequal (not (string= begstr endstr)))
         (begstr (cond
                  ((string= begstr "")
                   nil)
                  (regexp-quote (regexp-quote begstr))
                  (t begstr)))
         (endstr (cond ((string= endstr "")
                        nil)
                       (regexp-quote (regexp-quote endstr)) 
                       (t endstr)))
         (searchform (cond ((and beg-end-unequal begstr endstr)
                            (concat "\\(" begstr "\\)\\|\\(" endstr "\\)"))
                           (begstr begstr)
                           (endstr endstr)))
         (orig (point)) 
         beg-pos-delimiter end-pos-delimiter)
    (goto-char beg)
    (while
        (and
         (< (point) end) (not (eobp))
         (re-search-forward searchform end t 1))
      (if
          (or (and beg-end-unequal begstr (string-match begstr (match-string-no-properties 0)))
              (not beg-end-unequal))
          (if comment
              (unless (and condition (funcall condition))(setq nesting (1+ nesting)))
            (unless (or (and condition (funcall condition))(ar-in-comment-p-lor))
              (setq nesting (1+ nesting))))
        (when (string-match endstr (match-string-no-properties 0))
          (if comment
              (unless (and condition (funcall condition))(setq nesting (1- nesting)))
            (unless (or (and condition (funcall condition))(ar-in-comment-p-lor))
              (setq nesting (1- nesting)))))))
    (unless beg-end-unequal (setq nesting (% nesting 2)))
    (goto-char orig)
    nesting))

(provide 'beg-end)

;;; beg-end.el ends here
