;;; misc-utils.el --- Strip CHARS from STRING and more

;; Author: Andreas Roehler <andreas.roehler@easy-emacs.de>, unless indicated otherwise 

;; Keywords: convenience

;; This file is free software; you can redistribute it

;; and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2, or (at your option) any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Commentary: 
;; Provide some functions for convenience

;; Changes to previous version:

;; Moved forms like
;; `ar-skip-blanks-and-comments-forward-lor' from
;; ar-comment-lor into here

;;; History:
;; 

;;; Code:

(require 'string-strip)

(unless (boundp 'empty-line-p-chars) 
  (defcustom empty-line-p-chars "^[ \t\r]*$"
    "Empty-line-p-chars."
    :type 'regexp
    :group 'convenience))

(unless (functionp 'empty-line-p)
  (defun empty-line-p (&optional iact)
    "Returns t if cursor is at an empty line, nil otherwise."
    (interactive "p")
    (save-excursion
      (beginning-of-line)
      (when iact
        (message "%s" (looking-at empty-line-p-chars)))
      (looking-at empty-line-p-chars))))

;;;###autoload
(unless (functionp 'in-string-p)
  (defun in-string-p (&optional pos)
    (let ((orig (or pos (point))))
      (save-excursion
        (save-restriction
          (widen)
          (beginning-of-defun)
          (numberp (nth 3 (parse-partial-sexp (point) orig))))))))

(defun ar-in-string-p ()
  "Function `in-string-p' returns nil if at the first `\"'.
  Fixed here "
  (interactive)
  (let* ((erg (nth 8 (parse-partial-sexp (point-min) (point))))
         (la (unless erg (when (looking-at "\"")
                           (match-beginning 0)))))
    (setq erg (or erg la))
    (when (interactive-p) (message "%s" erg))
    erg))

(unless (functionp 'empty-stringp)
  (defun empty-stringp (string)
    (string= "" string)))

(unless (boundp 'empty-line-p-chars)
  (defcustom empty-line-p-chars "^[ \t\r]*$"
    "Empty-line-p-chars."
    :type 'regexp
    :group 'convenience))

;; redefining it here as existing function returs false positives
;; with GNU Emacs 23.1.50.1 (i686-pc-linux-gnu, GTK+ Version 2.12.0) of 2009-07-25 
(unless (featurep 'xemacs)
  (defun region-active-p ()
    "and mark-active transient-mark-mode
 (not (eq (region-beginning) (region-end)"
    (and mark-active transient-mark-mode
         (not (eq (condition-case nil (region-beginning)(error nil)) (condition-case nil (region-end) (error nil)))))))

(unless (functionp 'indent-with-spaces)
  (defun indent-with-spaces (column)
    "Indent with spaces to column "
    (interactive "*")
    (while (< (current-column) column)
      (insert " "))
    (forward-char 1)))

(when (featurep 'xemacs) 
  (unless (functionp 'looking-back)
    ;; from GNU Emacs subr.el
    (defun looking-back (regexp &optional limit greedy)
      "Return non-nil if text before point matches regular expression REGEXP.
    Like `looking-at' except matches before point, and is slower.
    LIMIT if non-nil speeds up the search by specifying a minimum
    starting position, to avoid checking matches that would start
    before LIMIT.
    If GREEDY is non-nil, extend the match backwards as far as possible,
    stopping when a single additional previous character cannot be part
    of a match for REGEXP."
      (let ((start (point))
            (pos
             (save-excursion
               (and (re-search-backward (concat "\\(?:" regexp "\\)\\=") limit t)
                    (point)))))
        (if (and greedy pos)
            (save-restriction
              (narrow-to-region (point-min) start)
              (while (and (> pos (point-min))
                          (save-excursion
                            (goto-char pos)
                            (backward-char 1)
                            (looking-at (concat "\\(?:"  regexp "\\)\\'"))))
                (setq pos (1- pos)))
              (save-excursion
                (goto-char pos)
                (looking-at (concat "\\(?:"  regexp "\\)\\'")))))
        (not (null pos))))))

(when (eq 0 (string-match "^23" emacs-version))
  (defalias 'region-exists-p 'use-region-p))

(defun buffer-narrowed-p ()
  (interactive) 
  "Returns a list, beg and end position of accessible buffer, 
if narrowed, nil otherwise. "
  (lexical-let* ((beg (point-min))
                 (end (point-max))
                 (erg (/= (buffer-size)
                          (- end
                             beg)))) 
    (when (interactive-p)
      (if erg
          (message "%s" (list beg end))
        (message "%s" erg)))
    erg))

(defun ar-current-indentation-position ()
  "Returns beginning position of code in line. "
  (interactive)
  (let ((here (point))
        (pos (progn (back-to-indentation)(point))))
    (prog1
        (point)
      (when (interactive-p) (message "%s" pos))
      (goto-char here))))

(defalias 'defun-beginning-position 'function-beginning-position) 
(defun function-beginning-position () 
  "Return the position where the current functions definition starts"
  (interactive)
  (save-excursion
    (let ((orig (point)) 
          (beg (when (beginning-of-defun) (point)))
          (end (when (end-of-defun) (point))))
      (if (and beg end (ignore-errors (<= orig end)))
          (progn
            (when (interactive-p) (message "%s" beg))
            beg)
        ;; if start was at beginning of function, end
        ;; of defun didn't succeed jumping back
        (goto-char orig)
        (when (end-of-defun) (setq end (point)))
        (when (beginning-of-defun) (setq beg (point)))
        (when (and beg end (ignore-errors (<= beg orig)))
          (when (interactive-p) (message "%s" beg))
          beg)))))

(defalias 'defun-end-position 'function-end-position)
(defun function-end-position () 
  "Print the position where the current functions definition ends"
  (interactive)
  (save-excursion
    (let ((orig (point))
          (end (when (end-of-defun) (point)))
          (beg (when (beginning-of-defun) (point))))
      (if (and beg end (ignore-errors (<= beg orig)))
          (progn
            (when (interactive-p) (message "%s" end))
            end)
        (goto-char orig)
        (when (beginning-of-defun) (setq beg (point)))
        (when (end-of-defun) (setq end (point)))
        (when (and beg end (<= orig end))
          (when (interactive-p) (message "%s" end))
          end)))))

(defalias 'copy-defun 'copy-function) 
(defun copy-function (&optional arg) 
  (interactive "p")
  (when (or
         (looking-at
          (if defun-prompt-regexp
              (concat "^\\s(\\|"
                      "\\(" defun-prompt-regexp "\\)\\s(")
            "^\\s("))
         (eq (current-column) 0))
    (end-of-line))
  (save-excursion
    (let ((start (if
                     (looking-at "(defun")
                     (point)
                   (defun-beginning-position arg)))
          (end (function-end-position arg)))
      (kill-new (buffer-substring-no-properties start end)))))

(defalias 'kill-defun 'kill-function)
(defun kill-function (&optional arg) 
  (interactive "p")
  (let ((arg (or arg 1)))
    (save-excursion
      ;; don't go backward, if already at beginning
      (when
          (or
           (looking-at (if defun-prompt-regexp
                           (concat "^\\s(\\|"
                                   "\\(" defun-prompt-regexp "\\)\\s(")
                         "^\\s("))
           (eq (current-column) 0))
        (end-of-line)) 
      (kill-region 
       (function-beginning-position arg) (function-end-position arg)))))

(defun ar-name-of-defun-atpt (&optional pos)
  "Return name of function definition at point. 
Goto optional arg POS, if given. "
  (interactive)
  (save-excursion
    (let (name)
      (when pos (goto-char pos))
      (beginning-of-defun)
      (forward-word 1)
      (when (looking-at " *\\([(]+\\) *")
        (setq name (match-string-no-properties 1)))
      (when (interactive-p) (message "%s" name))
      name)))

(defun goto-outbuf (outbuf &optional iact)
  "Go to the end of a named puffer, provide a newline.
Show it, if interactively called. " 
  (set-buffer outbuf)
  (when iact (switch-to-buffer (current-buffer)))
  (goto-char (point-max))
  (unless (empty-line-p) (newline)))

(defun ar-whitespaces-backward ()
  (interactive) 
  (skip-chars-backward " \t\r\n\f")
  (point))

(defun ar-whitespaces-forward ()
  (interactive) 
  (skip-chars-forward " \t\r\n\f")
  (point))

(defun ar-skip-blanks-and-comments-forward-lor (&optional arg) 
  "Go forward over empty lines and comments alike.
With negative arg go backward. "
  (interactive "p")
  (ar-skip-blanks-and-comments-lor arg))

(defun ar-skip-blanks-and-comments-backward-lor (&optional arg) 
  "Go forward over empty lines and comments alike.
With negative arg go backward. "
  (interactive "p")
  (ar-skip-blanks-and-comments-lor (if (< 0 arg) (- arg) arg)))

;; helpers
(defun ar-skip-blanks-and-comments-lor (&optional arg) 
  "Go forward over empty lines and comments alike.
With negative arg go backward. "
  (or arg (setq arg 1))
  (let ((pos (point))) 
    (if (< 0 arg)
        (progn 
          (skip-chars-forward " \t\r\n")
          (when (ar-in-comment-p-lor)
            (goto-char (ar-comment-end-position-atpt)))
          (when (empty-line-p)
            (forward-line arg)) 
          (when (> (point) pos)
            (ar-skip-blanks-and-comments-lor arg)))
      (skip-chars-backward " \t\r\n")
      (when (ar-in-comment-p-lor)
        (goto-char (ar-comment-beginning-position-atpt)))
      (when (empty-line-p)
        (forward-line arg)) 
      (when (< (point) pos)
        (ar-skip-blanks-and-comments-backward-lor arg)))))

(defun ar-skip-chars-to-comment-lor (arg)
  "With positive arg, skip forward until next comment.
  With negative arg, skip backward.
 If no comment found, beg or end of buffer is reached. "
  (if (< 0 arg)
      (while (not (or (ar-in-comment-p-lor)
                      ;; (nth 4 (parse-partial-sexp (point-min) (point)))
                      (eobp)))
        (forward-char 1))
    (while (not (or (ar-in-comment-p-lor)
                    ;; (nth 4 (parse-partial-sexp (point-min) (point)))
                    (bobp)))
      (forward-char -1))))

(defun ar-down-list (&optional arg)
  "Makes `down-list' work also if inside a string or comment. "
  (interactive "p")
  (while (or (when (in-string-p)
               (ar-forward-string-atpt)
               (forward-char 1))
             (when (ar-in-comment-p-lor)
               (ar-forward-comment-atpt)
               (forward-char 1))))
  (down-list arg))

(defun reverse-chars (&optional arg)
  "Reverse reciproke chars as \"[\" to \"]\", upcase or downcase,
Switch `\"' with `''
If over a number, add ARG to it.
With negative arg, substract from number. "
  (interactive "*p")
  (let (i) 
    (if (looking-at "[0-9]")
        (ar-add-to-number arg)
      (let* ((cf (char-after)) 
             (cn (downcase cf)))
        (cond ((or (eq cf 34)(eq cf ?\"))
               (setq cn "'"))
              ((or (eq cf 39)(eq cf ?\'))
               (setq cn "\""))
              ((or (eq cf 43)(eq cf ?\+))
               (setq cn "-"))
              ((or (eq cf 62)(eq cf ?\>))
               (setq cn "<"))
              ((or (eq cf 60)(eq cf ?\<))
               (setq cn ">"))
              ((or (eq cf 187)(eq cf ?\»))
               (setq cn "«"))
              ((or (eq cf 171)(eq cf ?\«))
               (setq cn "»")) 
              ((or (eq cf 41)(eq cf ?\)))
               (setq cn "("))
              ((or (eq cf 40)(eq cf ?\())
               (setq cn ")"))
              ((or (eq cf 123) (eq cf ?\{))
               (setq cn "}"))
              ((or (eq cf 125) (eq cf ?\}))
               (setq cn "{"))
              ((or (eq cf 93)(eq cf ?\]))
               (setq cn "["))
              ((or (eq cf 91)(eq cf ?\[))
               (setq cn "]"))
              ((or (eq cf 45)(eq cf ?\-))
               (setq cn "_"))
              ((or (eq cf 95)(eq cf ?\_))
               (setq cn "-"))
              ((or (eq cf 92)(eq cf ?\\))
               (setq cn "/"))
              ((or (eq cf 47)(eq cf ?\/))
               (setq cn "\\"))
              (t (when (eq cf cn)
                   (setq cn (upcase cf)))))
        (delete-char 1) 
        (insert cn)))))

(defun ar-switch-quotes (&optional beg end)
  "Replaces `'' with `\"' or `\"' with `'', first found decides. "
  (interactive)
  (lexical-let ((beg (cond (beg)
                           ((buffer-narrowed-p)
                            (point-min))
                           ((region-active-p)
                            (region-beginning))
                           (t (point-min))))
                (end (cond (end)
                           ((buffer-narrowed-p)
                            (point-max))
                           ((region-active-p)
                            (copy-marker (region-end)))
                           (t (point-max)))))
    (ar-switch-quotes-intern beg end)))

(defun ar-switch-quotes-intern (beg end)
  (goto-char beg)
  (let* ((oldquote (progn (re-search-forward "[\"']" end t 1) (goto-char (match-beginning 0))(char-after)))
         (newquote (if (ignore-errors (eq oldquote 34)) 39 34))
         erg)
    (while (and oldquote (looking-at (char-to-string oldquote)))
      (replace-match (char-to-string newquote))
      (setq erg (search-forward (char-to-string oldquote) end t 1))
      (when erg (goto-char (match-beginning 0))))))

;; * Simple useful functions
;; 
;; Date: Thu, 28 Oct 2010 11:56:15 -0700
;; From: Tak Ota <Takaaki.Ota@am.sony.com>
;; To: <emacs-devel@gnu.org>
;; Subject: simple useful functions
;; Envelope-To: andreas.roehler@online.de
;; 
;; If you think the following two functions are universally useful please
;; consider incorporating them in simple.el or any appropriate package.
;; If not disregard.
;; 
;; -Tak

(defun collect-string (regexp &optional num)
  "Collect strings of REGEXP (or optional NUM paren) from the
current buffer into a collection buffer."
  (interactive "sCollect string (regexp): \nP")
  (let ((collection-buffer
	 (get-buffer-create (format "*Collection of \"%s\" *" regexp))))
    (with-current-buffer collection-buffer (erase-buffer))
    (save-excursion
      (goto-char (point-min))
      (while (re-search-forward regexp nil t)
	(let ((str (match-string (or num 0))))
	  (if str
	      (with-current-buffer collection-buffer
		(insert str)
		(or (zerop (current-column))
		    (insert "\n")))))))
    (pop-to-buffer collection-buffer)
    (goto-char (point-min))))

;; -Tak
(defun source (script &optional shell keep-current-directory)
  "Source the specified shell script.
Source the shell SCRIPT and import the environment into this
emacs.  The optional SHELL specifies the shell other than the
default `shell-file-name'.  When KEEP-CURRENT-DIRECTORY is nil,
which is the default, the current directory is temporarily
changed to the directory where the script resides while sourcing
the script."
  (interactive "fscript file: ")
  (if (null shell)
      (setq shell shell-file-name))
  (with-temp-buffer
    (unless keep-current-directory
      (setq default-directory (file-name-directory script)))
    (call-process shell nil t nil "-c" (concat "source " script "; printenv"))
    (while (re-search-backward "^\\([^=]+\\)=\\(.*\\)$" nil t)
      (setenv (match-string 1) (match-string 2)))))

;;
(defun ar-add-to-number-and-forward (&optional arg)
  (interactive "p*") 
  (while (and (not (eobp))(looking-at "[0-9]"))
    (ar-add-to-number arg)
    (re-search-forward "[0-9]" nil t 1)))

(defun ar-add-to-number (arg &optional beg end)
  "Raise numbers found in buffer or region with count. "
  (interactive "p*")
  (let* ((beg1 (cond (beg)
                     ((region-active-p)
                      (region-beginning))))
         (end1 (cond (end)
                     ((region-active-p)
                      (copy-marker (region-end)))))
         (collect (eq 4 (prefix-numeric-value arg)))
         (arg (if (eq 4 (prefix-numeric-value arg))
                  1
                (or arg 1)))
         (count 1))
    (let* ((bounds (ar-bounds-of-number-atpt))
           (beg (car bounds))
           (end (cdr bounds)))
      (save-excursion 
        (if (and beg end)
            (ar-add-to-number-base arg (buffer-substring-no-properties beg end) beg end beg1 end1 count collect)
          (ar-add-to-number-intern arg beg end beg1 end1 count collect))))))

(defun ar-add-to-number-intern (arg beg end beg1 end1 count collect)
  (goto-char beg1)
  (while (and (< (point) end1)(re-search-forward "\\([0-9]+\\)" end1 t))
    (ar-add-to-number-base arg (match-string-no-properties 0) (match-beginning 0) (match-end 0) beg1 end1 count collect)))

(defun ar-add-to-number-base (arg zeichen beg end beg1 end1 count collect)
  (let ((zahl (if (< 0 arg)
                  (+ (string-to-number zeichen) count)
                (- (string-to-number zeichen) count)))
        pos)
    (delete-region beg end)
    (insert (format "%s" zahl))
    (when collect
      (if (< 0 arg)
          (setq count (+ count arg))
        (setq count (- count arg)))
      (when (< (setq pos (point)) end1)
        (ar-add-to-number-intern arg nil nil pos end1 count collect)))))

(defun ar-count-lines (&optional start end)
  "Count lines in accessible part of buffer.
If no optional arguments are given but buffer is narrowed,
use the narrowed part.
If no optional arguments are given but region is active,
use the region.
Else count from point-min until curser position - (point)

See http://debbugs.gnu.org/cgi/bugreport.cgi?bug=7115"
  (interactive)
  (lexical-let ((beg (cond (start)
                           ((buffer-narrowed-p)
                            (point-min))
                           ((region-active-p)
                            (region-beginning))
                           (t (point-min))))
                (end (cond (end)
                           ((buffer-narrowed-p)
                            (point-max))
                           ((region-active-p)
                            (copy-marker (region-end)))
                           (t (point))))
                erg)
    (setq erg (1+ (count-matches "[\n\C-m]" beg end)))
    (when (interactive-p) (message "%s" erg))
    erg))

(defvar ar-indentation-registered '()
  "Filled in by `ar-register-indentation' and used by `ar-apply-indentation'")

(defun ar-register-indentation (&optional beg end)
  "Registers current functions indentation into `ar-indentation-registered'.
 `ar-aplly-indentation' indents from current line according to this list "
  (interactive) 
  (let ((beg (cond (beg)
                   ((region-active-p)
                    (region-beginning))
                   ((bobp)
                    (point)) 
                   (t (defun-beginning-position)))) 
        (end (cond (end) 
                   ((region-active-p)
                    (copy-marker (region-end)))
                   ((eobp)
                    (point)) 
                   (t (defun-end-position))))
        indentlist)
    (ar-register-indentation-intern beg end)))

(defun ar-register-indentation-intern (beg end)
  (goto-char beg)
  (setq indentlist (cons (current-indentation) indentlist))
  (while (re-search-forward "^[ \t]+" end t)
    (setq indentlist (cons (current-indentation) indentlist))) 
  (setq ar-indentation-registered (nreverse indentlist))
  (message "%s" ar-indentation-registered)
  ar-indentation-registered)

(defun ar-apply-indentation (&optional beg)
  "Idents respective values in `ar-indentation-registered'"
  (interactive "*")
  (when beg (goto-char beg))
  (beginning-of-line)
  (dolist (elt ar-indentation-registered)
    (indent-to-column elt)
    (let ((orig (point))) 
      (delete-region orig (+ orig (skip-chars-forward " \t"))))
    (forward-line 1))) 

(provide 'misc-utils)
;;; misc-utils.el ends here
