;;; string-strip.el --- Remove leading and/or trailing spaces/chars

;; Copyright (C) 2010  Andreas Roehler

;; Author: Andreas Roehler <andreas.roehler@online.de>
;; Keywords: lisp

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;   Return a copy of STR with leading and/or trailing
;; spaces/chars removed.  If chars to remove are not
;; specified when calling, default values of
;; `strip-chars-before', `strip-chars-after' are used.

;; If no string delivered but region is active,
;; buffer-substring of region is replaced with the
;; result

;; Used a form resulting from a discussion at
;; http://lists.gnu.org/archive/html/emacs-devel/2006-06/msg00428.html
;; (string-match
;;  (concat "\\`\\([" s-c-b"]*\\)" s-c-p "\\([" s-c-a "]*\\)\\'") str)

;; A bug showed up, as the chars to remove from the ends
;; also were inside the string to preserve: Rewrite done below

;;; Code:

(defalias 'string-strip 'strip)
(defalias 'string-strip-right 'rstrip)
(defalias 'string-strip-left 'lstrip)

(defcustom strip-chars-before "[ \t\r\n]*"
  "Regexp indicating which chars shall be stripped before STRING. "
  
  :type 'string
  :group 'convenience)

;; (setq strip-chars-after "[ \t\r\n]*")
(defcustom strip-chars-after "[ \t\r\n]*\\'"
  "Regexp indicating which chars shall be stripped after STRING. "
  
  :type 'string
  :group 'convenience)

;;;###autoload
(defun rstrip (&optional str chars-after)
  "Return a string or change region with trailing characters
 removed. Defaults to the value uof
 `strip-chars-after', basically whitespace, which is customizable. 
    (rstrip \"spacious \") 
    ==> \"spacious\""
  (interactive "*")
  (lexical-let ((chars-after (if chars-after chars-after t))) 
    (strip str nil chars-after)))

(defun lstrip (&optional str chars-before)
  "Return a string or change region with leading characters
 removed. Defaults to the value uof
 `strip-chars-before', basically whitespace, which is customizable. 
    (lstrip \"     spacious\") 
    ==> \"spacious\""
  (interactive "*")
  (lexical-let ((chars-before (if chars-before chars-before t))) 
    (strip str chars-before)))

(defun strip (&optional str chars-before chars-after)
  "Return a copy of STR, leading and/or trailing CHARS removed.
Default of `CHARS-BEFORE', `CHARS-AFTER': space, tab, carriage return, newline, newpage. 
If no string is delivered but region is active, strip region. 
    (strip \"   spacious   \") 
    ==> \"spacious\""
  (interactive "*") 
  (let* ((got-string (stringp str))
         (beg (cond (got-string)
                    ((region-active-p)
                     (region-beginning))
                    (t (point-min)))) 
         (end (cond (got-string)
                    ((region-active-p)
                     (region-end))
                    (t (point-max))))
         (str (cond (got-string str)
                    (t 
                     (buffer-substring-no-properties beg end))))
         (s-c-b (cond ((stringp chars-before)
                       chars-before)
                      ((and (not chars-before) chars-after) "")
                      (t strip-chars-before)))
         (s-c-a (cond ((stringp chars-after)
                       chars-after)
                      ((and chars-before (not chars-after)) "")
                      (t strip-chars-after)))
         newstring erg)
    (save-excursion
      (with-temp-buffer
        (insert str)
        (unless (string= "" s-c-a)
          (skip-chars-backward s-c-a)
          (delete-region (point) (point-max)))
        (goto-char (point-min))
        (unless (string= "" s-c-b)
          (skip-chars-forward s-c-b)
          (delete-region (point-min) (point)))
        (setq erg (buffer-substring-no-properties (point-min) (point-max)))))
    (unless got-string (delete-region beg end)
            (insert erg))
    (when (interactive-p) (message "%s" erg))
    erg))



(provide 'string-strip)
;;; string-strip.el ends here

